@@ -138,8 +138,10 @@ static char *fsl = "\007";
 short mutt_ts_capability(void)
 {
   char *term = getenv("TERM");
-  char *tcaps;
+  const char *tcaps;
+#ifdef HAVE_USE_EXTENDED_NAMES
   int tcapi;
+#endif
   char **termp;
   char *known[] = {
     "color-xterm",
@@ -155,13 +157,13 @@ short mutt_ts_capability(void)
   };
 
   /* If tsl is set, then terminfo says that status lines work. */
-  tcaps = tigetstr("tsl");
+  tcaps = mutt_tigetstr ("tsl");
   if (tcaps && tcaps != (char *)-1 && *tcaps)
   {
     /* update the static defns of tsl/fsl from terminfo */
     tsl = safe_strdup(tcaps);
 
-    tcaps = tigetstr("fsl");
+    tcaps = mutt_tigetstr ("fsl");
     if (tcaps && tcaps != (char *)-1 && *tcaps)
       fsl = safe_strdup(tcaps);
 
@@ -172,16 +174,16 @@ short mutt_ts_capability(void)
   /* Beware: tigetflag returns -1 if XT is invalid or not a boolean. */
 #ifdef HAVE_USE_EXTENDED_NAMES
   use_extended_names (TRUE);
-  tcapi = tigetflag("XT");
+  tcapi = mutt_tigetflag ("XT");
   if (tcapi == 1)
     return 1;
 #endif /* HAVE_USE_EXTENDED_NAMES */
 
   /* Check term types that are known to support the standard escape without
    * necessarily asserting it in terminfo. */
-  for (termp = known; termp; termp++)
+  for (termp = known; *termp; termp++)
   {
-    if (term && *termp && mutt_strncasecmp (term, *termp, strlen(*termp)))
+    if (term && !mutt_strncasecmp (term, *termp, strlen(*termp)))
       return 1;
   }
 
@@ -323,16 +325,34 @@ static int ci_first_message (void)
     if (old != -1)
       return (old);
 
-    /* If Sort is reverse and not threaded, the latest message is first.
-     * If Sort is threaded, the latest message is first iff exactly one
-     * of Sort and SortAux are reverse.
+    /* If Sort is threaded, the latest message is first iff exactly one
+     * of Sort and the top-level sorting method are reverse.
      */
-    if (((Sort & SORT_REVERSE) && (Sort & SORT_MASK) != SORT_THREADS) ||
-	((Sort & SORT_MASK) == SORT_THREADS &&
-	 ((Sort ^ SortAux) & SORT_REVERSE)))
-      return 0;
+    if ((Sort & SORT_MASK) == SORT_THREADS)
+    {
+      if ((SortThreadGroups & SORT_MASK) == SORT_AUX)
+      {
+        if ((Sort ^ SortAux) & SORT_REVERSE)
+          return 0;
+        else
+          return (Context->vcount ? Context->vcount - 1 : 0);
+      }
+      else
+      {
+        if ((Sort ^ SortThreadGroups) & SORT_REVERSE)
+          return 0;
+        else
+          return (Context->vcount ? Context->vcount - 1 : 0);
+      }
+    }
+    /* If Sort is reverse and not threaded, the latest message is first. */
     else
-      return (Context->vcount ? Context->vcount - 1 : 0);
+    {
+      if (Sort & SORT_REVERSE)
+        return 0;
+      else
+        return (Context->vcount ? Context->vcount - 1 : 0);
+    }
   }
   return 0;
 }
@@ -764,6 +784,25 @@ int mutt_index_menu (void)
 
     if (!in_pager)
     {
+
+#if defined (USE_SLANG_CURSES) || defined (HAVE_RESIZETERM)
+      while (SigWinch)
+      {
+        do
+        {
+          SigWinch = 0;
+          mutt_resize_screen ();
+        }
+        while (SigWinch);
+
+	/*
+	 * force a real complete redraw.  clrtobot() doesn't seem to be able
+	 * to handle every case without this.
+	 */
+	clearok(stdscr,TRUE);
+      }
+#endif
+
       index_menu_redraw (menu);
 
       /* give visual indication that the next command is a tag- command */
@@ -787,21 +826,6 @@ int mutt_index_menu (void)
                           MuttIndexWindow->cols - 1);
       mutt_refresh ();
 
-#if defined (USE_SLANG_CURSES) || defined (HAVE_RESIZETERM)
-      if (SigWinch)
-      {
-	SigWinch = 0;
-	mutt_resize_screen ();
-	menu->top = 0; /* so we scroll the right amount */
-	/*
-	 * force a real complete redraw.  clrtobot() doesn't seem to be able
-	 * to handle every case without this.
-	 */
-	clearok(stdscr,TRUE);
-	continue;
-      }
-#endif
-
       op = km_dokey (MENU_MAIN);
 
       dprint(4, (debugfile, "mutt_index_menu[%d]: Got op %d\n", __LINE__, op));
@@ -1071,8 +1095,9 @@ int mutt_index_menu (void)
                  Prompt when trying to quit Mutt while there are backgrounded
                  compose sessions in process.
               */
-              mutt_yesorno (_("There are $background_edit sessions. Really quit Mutt?"),
-                            MUTT_NO) != MUTT_YES)
+              mutt_query_boolean (OPTBACKGROUNDCONFIRMQUIT,
+                  _("There are $background_edit sessions. Really quit Mutt?"),
+                  MUTT_NO) != MUTT_YES)
           {
             break;
           }
@@ -1446,6 +1471,7 @@ int mutt_index_menu (void)
         break;
       }
 
+      case OP_GENERIC_SELECT_ENTRY:
       case OP_DISPLAY_MESSAGE:
       case OP_DISPLAY_HEADERS: /* don't weed the headers */
 
@@ -1499,8 +1525,9 @@ int mutt_index_menu (void)
 	{
           if (mutt_background_has_backgrounded () &&
               option (OPTBACKGROUNDCONFIRMQUIT) &&
-              mutt_yesorno (_("There are $background_edit sessions. Really quit Mutt?"),
-                            MUTT_NO) != MUTT_YES)
+              mutt_query_boolean (OPTBACKGROUNDCONFIRMQUIT,
+                  _("There are $background_edit sessions. Really quit Mutt?"),
+                  MUTT_NO) != MUTT_YES)
           {
             break;
           }
@@ -2635,9 +2662,10 @@ int mutt_index_menu (void)
         break;
 #endif
 
-      default:
+      case OP_NULL:
 	if (!in_pager)
 	  km_error_key (MENU_MAIN);
+        break;
     }
 
     if (in_pager)
