@@ -1050,10 +1050,15 @@ static int imap_open_mailbox_append (CONTEXT *ctx, int flags)
   if (rc == -1)
     return -1;
 
-  snprintf (buf, sizeof (buf), _("Create %s?"), mailbox);
-  if (option (OPTCONFIRMCREATE) &&
-      mutt_query_boolean (OPTCONFIRMCREATE, buf, 1) < 1)
-    return -1;
+  if (option (OPTCONFIRMCREATE))
+  {
+    if (option (OPTNOCURSES))
+      return -1;
+
+    snprintf (buf, sizeof (buf), _("Create %s?"), mailbox);
+    if (mutt_query_boolean (OPTCONFIRMCREATE, buf, 1) < 1)
+      return -1;
+  }
 
   if (imap_create_mailbox (idata, mailbox) < 0)
     return -1;
@@ -2367,7 +2372,7 @@ imap_complete_hosts (char *dest, size_t len)
     if (conn->account.type != MUTT_ACCT_TYPE_IMAP)
       continue;
 
-    mutt_account_tourl (&conn->account, &url);
+    mutt_account_tourl (&conn->account, &url, 0);
     /* FIXME: how to handle multiple users on the same host? */
     url.user = NULL;
     url.path = NULL;
@@ -2396,7 +2401,7 @@ int imap_complete(char* dest, size_t dlen, const char* path)
   char buf[LONG_STRING*2];
   IMAP_LIST listresp;
   char completion[LONG_STRING];
-  int clen;
+  size_t clen;
   size_t matchlen = 0;
   int completions = 0;
   IMAP_MBOX mx;
