@@ -1,7 +1,7 @@
 /* Copyright (C) 2004 Justin Hibbits <jrh29@po.cwru.edu>
  * Copyright (C) 2004 Thomer M. Gil <mutt@thomer.com>
  * Copyright (C) 2015-2016 Richard Russon <rich@flatcap.org>
- * Copyright (C) 2016-2017 Kevin J. McCarthy <kevin@8t8.us>
+ * Copyright (C) 2016-2017,2020-2021 Kevin J. McCarthy <kevin@8t8.us>
  *
  *     This program is free software; you can redistribute it and/or modify
  *     it under the terms of the GNU General Public License as published by
@@ -258,14 +258,15 @@ static int cb_qsort_sbe (const void *a, const void *b)
 
   switch ((SidebarSortMethod & SORT_MASK))
   {
+    /* Note: the three numeric counts below are reversed on purpose. */
     case SORT_COUNT:
-      result = (b2->msg_count - b1->msg_count);
+      result = mutt_numeric_cmp (b2->msg_count, b1->msg_count);
       break;
     case SORT_UNREAD:
-      result = (b2->msg_unread - b1->msg_unread);
+      result = mutt_numeric_cmp (b2->msg_unread, b1->msg_unread);
       break;
     case SORT_FLAGGED:
-      result = (b2->msg_flagged - b1->msg_flagged);
+      result = mutt_numeric_cmp (b2->msg_flagged, b1->msg_flagged);
       break;
     case SORT_PATH:
       result = mutt_strcasecmp (mutt_b2s (b1->pathbuf), mutt_b2s (b2->pathbuf));
