@@ -105,14 +105,8 @@ int _mutt_system (const char *cmd, int flags)
       }
     }
 
-    /* reset signals for the child; not really needed, but... */
     mutt_unblock_signals_system (0);
-    act.sa_handler = SIG_DFL;
-    act.sa_flags = 0;
-    sigemptyset (&act.sa_mask);
-    sigaction (SIGTERM, &act, NULL);
-    sigaction (SIGTSTP, &act, NULL);
-    sigaction (SIGCONT, &act, NULL);
+    mutt_reset_child_signals ();
 
     execle (EXECSHELL, "sh", "-c", cmd, NULL, mutt_envlist ());
     _exit (127); /* execl error */
